from Products.CMFCore.utils import UniqueObject 
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo, Unauthorized
from Products.CMFCore import CMFCorePermissions
import zLOG,os
from Products.Archetypes.utils import DisplayList
from Products.CMFCore.utils import getToolByName
from Products.Annotations.AnnotationTool import getAnnotations
import Persistence
from ZODB.PersistentMapping import PersistentMapping
from Products.CMFCore.CMFCorePermissions import setDefaultRoles

from Permissions import MANAGE_VOCABULARIES

setDefaultRoles(MANAGE_VOCABULARIES, ('Manager',))

EXTEND_VOCAB=0
REPLACE_VOCAB=1
ID_TYPE_STRING=0
ID_TYPE_INT = 1
ORDERBY_ID=1
ORDERBY_TITLE=2
ORDER_NONE=0
EDITABLE_FULL=2
EDITABLE_EXT_ONLY=1
EDITABLE_NONE=0

SimpleVocabularyNamespace = 'http://www.companion.nl/xml/simplevocabularies'

def isNumeric(str):
    try:
        i = int(str)
        return True
    except ValueError:
        return False

#============================ 
class SimpleVocabulariesContainer(SimpleItem, Persistence.Persistent):
    """ container that holds vocabularies """
    
    def __init__(self):
        self.vocabs = {}

    def __len__(self):
        return len(self.vocabs)

    def getVocabularyNames(self):
        return self.vocabs.keys()
    
    def getVocabulary(self, name):
        if self.vocabs.has_key(name):
            return self.vocabs[name]
        else:
            raise AttributeError('Vocabulary not found')
        
    def hasVocabulary(self, name):
        return self.vocabs.has_key(name)
    
    def addVocabulary(self, vocabulary):
        self.vocabs[vocabulary.name]=vocabulary
        self._p_changed = 1
        
    def removeVocabulary(self, name):
        if self.hasVocabulary(name):
            del self.vocabs[name]
            self._p_changed = 1

    def hasCustomizations(self):
        if len(self.vocabs)>0:
            return True
        else:
            return False
            
                
#============================        
class SimpleVocabulary(DisplayList, SimpleItem):
    """  vocabulary class """
    
    def __init__(self, name, description='', mode=EXTEND_VOCAB, editable=EDITABLE_FULL, data=None, idType=ID_TYPE_STRING, orderBy=ORDERBY_TITLE):
        self.name=name
        self.description=description
        self.mode=mode
        self.editable=editable
        self.idType=idType
        self.orderBy=orderBy  # just a flag telling the tool that end result for this vocabulary needs to be sorted after traversal.
        if data!=((),) and data:
            DisplayList.__init__(self, data)
        else:
            DisplayList.__init__(self)
        
        #SimpleVocabulary.inheritedAttribute('__init__')(self, data)
        
    def resetVocab(self, data):
        DisplayList.__init__(self, data)
        
    def getValues(self):
        """ return values controlled by the orderBy attribute """
        
        if self.orderBy==ORDER_NONE:
            return self
        elif self.orderBy==ORDERBY_ID:
            return self.sortedByKey()
        else:
            return self.sortedByValue()
        
    def getValuesForFormulator(self):
        """ Return the vocabulary formatted for formulator """
        
        # formulator builds the list [ (title, id), (title, id) ... ]
        # exactly the other way round from what a DisplayList will return

        def _cmpkey(a, b):
                return cmp(a[1], b[1])        

        def _cmpvalue(a, b):
            return cmp(a[0], b[0])
        
        keys = self.keys()
        data = [(self.getValue(key), key) for key in keys]
        
        if self.orderBy==ORDER_NONE:
            return tuple(data)
        elif self.orderBy==ORDERBY_ID:
            data.sort(_cmpkey)
            return tuple(data)
        else:
            data.sort(_cmpvalue)
            return tuple(data)
        
    def getFormattedValues(self):
        values = self.getValues()
        items = values.items()
        result=[]
        for key, value in items:
            if key==value:
                result.append(value)
            else:
                result.append(str(key) + '|' + str(value))
        return result
        
    def keysAreNumeric(self): 
        keys = self.keys()
        for key in keys:
            if not isNumeric(key):
                return False
        return True
            
        

        
#============================        
class SimpleVocabulariesTool(UniqueObject, SimpleItem,PropertyManager): 
    """ This tool manages vocabularies """ 
    id = 'portal_vocabularies' 
    meta_type= 'SimpleVocabularies tool' 
    plone_tool = 1
        
    manage_options=PropertyManager.manage_options
    
    security = ClassSecurityInfo()

    # create a container that holds all the portal vocabularies
    # when one wants to add a vocabulary then it needs to 
    # exist inside this container
       
    #=====================================================    
    #=====================================================

    def __init__(self):
        self.portalVocabs = SimpleVocabulariesContainer() 
        
        # add properties
        self.manage_addProperty('limitToFolderishTypes', 0, 'boolean')
        self.manage_addProperty('typesAllowModification', '', 'lines')
        self.manage_addProperty('typesNotAllowModification', '', 'lines')

    #=====================================================    
    #=====================================================
        
    def __len__(self):
        return len(self.portalVocabs)

    #=====================================================    
    #=====================================================

    def _processObject(self, object, param_names):
        # get the container
        container = self.getLocalVocabularyContainer(object)
        if container:
            # get the names 
            names = param_names.split('||')[0].split('\n')

            print 'Removing vocabularies ' +  ',  '.join(names) + ' from :' + object.title_or_id()
            
            for name in names:
                container.removeVocabulary(name)

    #=====================================================    
    #=====================================================

    def processVocabularyData(self, vocab_data, idType=ID_TYPE_STRING):
        # ( (id1, term1), (id2, term2), .. , (idn, termn) )
        data = ()
        keys=[]
        for term in vocab_data:
            if term!='':
                if term.find('|')>0:
                    # user defined id
                    splitted = tuple([t.strip() for t in term.split('|')])
                    if len(splitted)>2:
                        raise SyntaxError, 'Malformed tems. Only one | is allowed per term.'
                    
                    key, value = tuple([t.strip() for t in term.split('|')])

                    if idType==ID_TYPE_INT:
                        try:
                            key = int(key)
                        except ValueError:
                            raise SyntaxError, 'Ids are not numeric'
                     
                    s_term = (key, value)
                    if not key in keys:
                        data = data + (s_term,)
                        keys.append(key)
                    else:
                        raise ValueError, 'ID \'' + key + '\' is not unique'
                else:
                    # generate an id automatically
                    if idType==ID_TYPE_INT:
                        raise Exception, 'Ids need to be provided in combination with integer ID type.'
                    if not term in keys:
                        data = data + ((term, term),) 
                        keys.append(term)
                    else:
                        raise ValueError, 'Term \'' + term + '\' is not unique'
                    
        if data==():
            data=((),)
            
        return data

    #=====================================================    
    #=====================================================

    def addVocabulary(self, name, context, vocabulary):
        """ add a vocabulary to the current context """
        
        portalObject = context.portal_url.getPortalObject()
        
        # if context is the portal then add it to the tool instead of the object.
        if context == portalObject:
            self.addPortalVocabulary(name, vocabulary)
        else:
            annotation = getAnnotations(context, SimpleVocabularyNamespace)
    
            # create the container if the context doesn't have one yet
            if not annotation.has_key('SimpleVocabulariesContainer'):
                container = annotation['SimpleVocabulariesContainer']=SimpleVocabulariesContainer()
            else:
                container = annotation['SimpleVocabulariesContainer']
                
            # add the vocabulary, overwrite if it already exists
            container.addVocabulary(vocabulary)
            # reindex this object to make sure that our indexes are up to date.
            self.portal_catalog.reindexObject(object = context, idxs=['vocabulary_customizations','hasVocubularyCustomizations'])

    #=====================================================    
    #=====================================================
    
    def getPortalVocabulary(self, name):
        """ get the tool's/portal's vocabulary """
        if self.hasVocabulary(name):
            return self.portalVocabs.getVocabulary(name)
        else:
            raise AttributeError('Vocabulary not found')

    #=====================================================    
    #=====================================================
    
    def addPortalVocabulary(self, vocabulary):
        """ add a vocabulary to the tool and therefore the portal """
        self.portalVocabs.addVocabulary(vocabulary)

        
    #=====================================================    
    #=====================================================

    
    def getLocalVocabularyContainer(self, object):
        annotation = getAnnotations(object, SimpleVocabularyNamespace)
        # create the container if the context doesn't have one yet
        if annotation.has_key('SimpleVocabulariesContainer'):
            return annotation['SimpleVocabulariesContainer']
        else:
            return None

    #=====================================================    
    #=====================================================

    def getLocalVocabularyNames(self, object):
        """ return the names for the customizations for a given object """

        if object == self.portal_url.getPortalObject() or object == self:
            return []
                
        annotation = getAnnotations(object, SimpleVocabularyNamespace)
        # create the container if the context doesn't have one yet
        if annotation.has_key('SimpleVocabulariesContainer'):
            container = annotation['SimpleVocabulariesContainer']
            names = container.getVocabularyNames()
            return names
        else:
            return []
        
    #=====================================================    
    #=====================================================
        
    def getLocalVocabularyObject(self, name, object):
        """ return the local vocabulary object for object """
        annotation = getAnnotations(object, SimpleVocabularyNamespace)
        vocabularyContainer = annotation.get('SimpleVocabulariesContainer', None)
        if vocabularyContainer:
            if vocabularyContainer.hasVocabulary(name):
                return vocabularyContainer.getVocabulary(name)
            else:
                return None
        else:
            return None

    #=====================================================    
    #=====================================================
    
    def getVocabularies(self):
        return self.portalVocabs

    #=====================================================    
    #=====================================================
        
    def getVocabulary(self, name, context, acquiredOnly=False, formulator=False):
        """ return the actual vocabulary (terms) in a given context """
        # set acquiredOnly to True if you are only interested in the acquired vocabulary, not including
        # set formulator to True if you want to get the list formatted for formulator widgets
        # the vocabulary of context itself.
        
        # traverse the path and determine the vocabulary's terms
        
        # first check if the vocab exist
        if not self.hasVocabulary(name):
            raise AttributeError('Vocabulary not found')
        
        toolVocab = self.getPortalVocabulary(name)
        portalObject = context.portal_url.getPortalObject()
        
        # now check if the vocabulary is set to globalOnly, do a shortcut in that case
        if toolVocab.editable==EDITABLE_NONE or context == portalObject:
            if formulator:
                return self.portalVocabs.getVocabulary(name).getValuesForFormulator()
            else:
                return self.portalVocabs.getVocabulary(name).getValues()
        
        # now do the traversal

        
        if acquiredOnly:
            parent = context.aq_parent
        else:
            parent = context
            
        currentVocab=None
        
        result=None
        
        while parent != portalObject:

            #if formulator:
                #localVocab = localVocab.getValuesForFormulator()
            #else:
            # get the local vocabulary object if it exists
            localVocab = self.getLocalVocabularyObject(name, parent)
            
            if localVocab:
                # add localVocab's vocabulary 
                if not currentVocab:
                    currentVocab = localVocab
                else:
                    currentVocab = localVocab + currentVocab
                # check if we have to keep on traversing or if this is the end-point.
                # check for EDITABLE_FULL because it could have happened that
                # toolVocab.editable was set from FULL to EXTEND_ONLY
                # in that case treat localVocab.mode as EXTEND_VOCAB and keep traversing upwards
                if localVocab.mode==REPLACE_VOCAB and toolVocab.editable==EDITABLE_FULL:
                    result = currentVocab
                    break

            parent = parent.aq_parent
                
         # the portal doesn't have a SimpleVocabulariesContainer, instead
         # we return the SimpleVocabulariesContainer from the tool itself
        if not result: 
            if not currentVocab:
                result = toolVocab
            else:
                result = currentVocab +toolVocab
         
        if formulator:
            #return toolVocab.getValuesForFormulator()
            return self.sortForFomulator(result, toolVocab.orderBy)
        else:
             # result is at least a DisplayList
             # sort the stuff
            if toolVocab.orderBy==ORDER_NONE:
                return result
            elif toolVocab.orderBy==ORDERBY_ID:
                return result.sortedByKey()
            else:
                return result.sortedByValue()             
        
    #=====================================================    
    #=====================================================

    def getVocabularyFormatted(self, name, context, acquiredOnly=False, formulator=False):
        values = self.getVocabulary(name, context, acquiredOnly, formulator)
        items = values.items()
        result=[]
        for key, value in items:
            if key==value:
                result.append(value)
            else:
                result.append(str(key) + '|' + str(value))        
        return result

    #=====================================================    
    #=====================================================

    def getFieldFromRequest(self, request, index, field):
        voc = request.get('voc', '')
        if voc:
            return voc[index][field]
        else:
            return None
    
    #=====================================================    
    #=====================================================
    def getVocabularyNames(self):
        return self.portalVocabs.getVocabularyNames()

    #=====================================================    
    #=====================================================

    def getVocabularyCustomizations(self, names, sortbypath=0):
        """ returns all the customizations (object, container) done on objects
             for further processing or validations """

        pc = self.portal_catalog
        query = {'vocabulary_customizations': names}
        brains = pc.searchResults(query)
        result=[]
        for b in brains:
            object = b.getObject()
            container = self.getLocalVocabularyContainer(object)
            if container:
                result.append((object, container))
        if sortbypath:
            result.sort(lambda x,y:cmp(x[0].absolute_url(), y[0].absolute_url()))
            return result
        else:
            return result


    #=====================================================    
    #=====================================================
            
    def hasVocabulary(self, name):
        return self.portalVocabs.hasVocabulary(name)

    #=====================================================    
    #=====================================================
            
    def hasVocabularyCustomizations(self, object):
        """ indicator to see if an object has a customization for a vocabulary
             used for the catalog 
        """
        if object == self.portal_url.getPortalObject() or object == self:
            return 0
        
        container  = self.getLocalVocabularyContainer(object)
        if container.hasCustomizations():
            return 1
        else:
            return 0


    #=====================================================    
    #=====================================================
        
    def manage_AddPortalVocabulary(self, REQUEST=None):
        """ proces add vocab from form"""
        if REQUEST==None:
            return False
        
        # assume validation has been done already
        name = REQUEST.get('new_name')
        description = REQUEST.get('new_description')
        editable = REQUEST.get('new_editable', EDITABLE_FULL)
        idType = REQUEST.get('new_idType', ID_TYPE_STRING)
        orderBy=REQUEST.get('new_orderBy', ORDERBY_TITLE)
        
        vocabulary = REQUEST.get('new_vocabulary')
        
        # create the vocabulary
        try:
            data = self.processVocabularyData(vocabulary, idType)
        except ValueError:
            return False

        voc = SimpleVocabulary(name=name, description=description, editable=editable,  data=data, idType=idType, orderBy=orderBy)
        self.addPortalVocabulary(vocabulary = voc)

        return True

    #=====================================================    
    #=====================================================
    
    def manage_DeletePortalVocabulary(self, REQUEST=None):
        """ proces delete vocab from form"""

        if REQUEST==None:
            return  'Removed 0 vocabularies'

        data = REQUEST['voc']
        removed = []
        for voc in data:
            if voc.has_key('delete'):
                # delete the vocabulary
                self.removePortalVocabulary(voc['name'], process_all=False)  #process_all later in one go to speed up.
                removed.append(voc['name'])

        if removed:
            self.removeAllVocabularies(removed)
                
        if removed:
            return 'Removed ' + ', '.join(removed)
        else:
            return 'Removed 0 vocabularies'

    #=====================================================    
    #=====================================================
       
    def manage_SavePortalVocabulary(self, REQUEST=None):
        """ proces save vocab from form"""
        if REQUEST==None:
            return  'Nothing saved.'

        data = REQUEST['voc']
        
        for voc in data:
            if self.hasVocabulary(voc['name']):
                # first get the existing instance
                name = voc['name']
                vocabulary=self.getPortalVocabulary(voc['name'])
                vocabulary.description=voc['description']

                vocabulary.editable=voc['editable-' + name]
                # todo: when the permission changes then we may have
                # to modify the existing customization:
                # e.g. if editiable has changed from FULL to EXTEND_ONLY
                # then all REPLACE_VOCABs need to be set to EXTEND_ONLY
                # or is it enough to just ignore this REPLACE and treat them
                # like EXTEND when the vocabuluary is requested?
                
                vocabulary.idType = voc['idType-' + name]
                # todo: what happens if this type changes?
                # it could invalidate all the ids in the customizations
                
                vocabulary.orderBy = voc['orderBy-' + name]
                try:
                    vocdata=self.processVocabularyData(voc['vocabulary'], vocabulary.idType)
                except SyntaxError, msg:
                    return False, name, msg
                except Exception, msg:      
                    return False, name, msg
                except ValueError, msg:
                    return False, name, msg
                # id's within the passed vocabulary set is unique, 
                # now check if they are unique site-wide
                # of course you can never be sure that the keys are always unique
                # when you copy an object with a customization then ids are duplicated.
                # i consider that the user's own risk for now
                
                # for now I leave it as it is
                vocabulary.resetVocab(vocdata)
        
        return True, 'Vocabulary changes are saved.', ''

    #=====================================================    
    #=====================================================
    
    def manage_SaveObjectVocabulary(self, context,  REQUEST=None):
        """ proces save local vocab from form"""
        if REQUEST==None:
            return  'Nothing saved.'

        data = REQUEST['voc']
        
        for voc in data:
            name = voc['name']
            if voc['editable']!=EDITABLE_NONE:
                # assume the UI deals with EDITABLE settings.
                if voc['mode-' + name]==EXTEND_VOCAB and voc['vocabulary']==[]:
                    # remove the local vocabulary
                    self.removeVocabulary(context, voc['name'])
                else:
                    # simply make a new vocabulary object
                    idType = voc['idType']

                    try:
                        vocdata=self.processVocabularyData(voc['vocabulary'], idType=idType)
                    except SyntaxError, msg:
                        return False, name, msg
                    except Exception, msg:      
                        return False, name, msg
                    except ValueError, msg:
                        return False, name, msg                
                    mode = voc['mode-' + name]
                    vocabulary=SimpleVocabulary(name=name, mode=mode, data=vocdata)
                    # add this one
                    self.addVocabulary(name, context, vocabulary)
            
        return True, 'Vocabulary changes saved.', ''

    
    #=====================================================    
    #=====================================================

    def removeAllVocabularies(self, names):
        """ remove all folder/object customizations of a given vocabulary """

        tuples = self.getVocabularyCustomizations(names)
        pc = self.portal_catalog

        for object, container in tuples:
            for name in names:
                container.removeVocabulary(name)
                pc.reindexObject(object = object, idxs=['vocabulary_customizations','hasVocubularyCustomizations'])
    
    
    def removeAllVocabulariesOld(self, names):
        """ remove all folder/object customizations of a given vocabulary """
        
        # use a trick to pass names to _processObject.
        # first we glue the names together with a delimiter and then  add || to it so we can 
        # remove the stuff that Zope adds later.
        param = '\n'.join(names) + '||'
        portalObject = self.portal_url.getPortalObject()
        self.ZopeFindAndApply(portalObject, search_sub=1, apply_func=self._processObject, apply_path=param)
            
    #=====================================================    
    #=====================================================
   
    def removePortalVocabulary(self, name, process_all=False):
        if self.hasVocabulary(name):
            self.portalVocabs.removeVocabulary(name)
            if process_all:
                self.removeAllVocabularies([name])
                
    #=====================================================    
    #=====================================================

    def removeVocabulary(self, context, name, process_all=False):
        """ remove the vocabulary with id from context's SimpleVocabulariesContainer """

        portalObject = context.portal_url.getPortalObject()
        
        # if context is the portal then remove it from the tool instead of the object.
        if context == portalObject:
            self.removePortalVocabulary(name, process_all)
        else:
            container=self.getLocalVocabularyContainer(context)
            if container:
                container.removeVocabulary(name)

    #=====================================================    
    #=====================================================
    
    def modificationAllowed(self, object):
        """ function used by portal_actions to see if the tab needs to be shown. """
        allow = False
        
        if self.getProperty('limitToFolderishTypes', 0)==1:
            if object.isPrincipiaFolderish:
                allow = True
            else:
                allow = False
        else:
            allow=True

        type = object.portal_type
        allowed_types = self.getProperty('typesAllowModification', '')
        not_allowed_types = self.getProperty('typesNotAllowModification', '')
        
        if not  (allowed_types=='' or allowed_types==[] or allowed_types==()):
            if not type in allowed_types:
                allow = False

        if not  (not_allowed_types=='' or not_allowed_types==[] or not_allowed_types==()):
            if type in not_allowed_types:
                allow = False
        
        return allow
        
    #=====================================================    
    #=====================================================
    
    def sortForFomulator(self, data, orderBy):
        # formulator builds the list [ (title, id), (title, id) ... ]
        # exactly the other way round from what a DisplayList will return

        def _cmpkey(a, b):
                return cmp(a[1], b[1])        

        def _cmpvalue(a, b):
            return cmp(a[0], b[0])
        
        keys = data.keys()
        new_data = [(data.getValue(key), key) for key in keys]        

        if orderBy==ORDER_NONE:
            return tuple(new_data)
        elif orderBy==ORDERBY_ID:
            new_data.sort(_cmpkey)
            return tuple(new_data)
        else:
            new_data.sort(_cmpvalue)
            return tuple(new_data)        
        
    #=====================================================    
    #=====================================================
    

    def verifyIDTypeTransition(self, name):
        """ Check if a vocabulary may be set from string to int id type 
             A transition is not allowed when:
             customizations exist that have vocabs with terms without ids
             customizations exist that have ids that are not numeric
        """ 
        customizations = self.getVocabularyCustomizations([name])
        for object, container in customizations:
           vocab = container.getVocabulary(name)
           if not vocab.keysAreNumeric():
               return False  #transition not allowed
           
        return True
        
     
InitializeClass(SimpleVocabulariesTool)

        